// bplc Coroutine.bpl generators.bpl

let

open System.List

val make = \ (x) (xs) => Cons ( head = x, tail = xs )

function revto <a> (front, rear) : List <a> =
  case front of
    Nil       => rear
  | Cons cons => revto <a> (cons.tail, make cons.head rear)
  end

type Gen <a> = () -> Maybe <a>
data Maybe <a> = Nothing | Just a

function foreach <a> (gen : Gen <a>) (action : a -> ()): () =
  case gen () of
    Nothing => ()
  | Just x  => action x; foreach <a> gen action
  end

///////////////////////////////////////////////////////////////////////////////

// elements of a list

// state

function elements <a> (root : List <a>) : Gen <a> =
  Coroutine.define (\ yield =>
    fun () =>
      System.List.foreach <a> root (\ elem => yield (Just elem));
      Nothing
  )

// coroutine

function elements' <a> (root : List <a>) : Gen <a> =
  let val xs = ref root
  in
    fun () =>
      case !xs of
        Nil       => Nothing
      | Cons cons => xs := cons.tail; Just (cons.head)
      end
  end

///////////////////////////////////////////////////////////////////////////////

function sublists <a> (root : List <a>) : Gen <List <a>> =
  Coroutine.define (\ yield =>
    fun () =>
      case root of
        Nil       => yield (Just Nil)
      | Cons cons => foreach <List <a>> (sublists <a> (cons.tail)) (\ xs =>
                       yield (Just xs);
                       yield (Just (make cons.head xs))
                     )
      end;
      Nothing
  )

///////////////////////////////////////////////////////////////////////////////

function permutations <a> (root : List <a>) : Gen <List <a>> =
  Coroutine.define (\ yield =>
    fun () =>
      case root of
        Nil       => yield (Just Nil)
      | Cons cons => foreach <List <a>> (permutations <a> (cons.tail)) (\ xs =>
                       let function insertions (left, right) : () =
                             yield (Just (revto <a> (left, make cons.head right)));
                             case right of
                               Nil       => ()
                             | Cons cons => insertions (make cons.head left, cons.tail)
                             end
                       in
                         insertions (Nil, xs)
                       end
                     )
      end;
      Nothing
  )

///////////////////////////////////////////////////////////////////////////////

val n = ref 0

in

// System.IO.println (revto <Nat> (Nil, fromto (0, 9))); 

//System.List.foreach <Nat> (fromto (0, 9)) (\ x =>
//foreach <Nat> (elements' <Nat> (fromto (0, 9))) (\ x =>
//foreach <Nat> (sublists <Nat> (fromto (0, 3))) (\ x =>
foreach <List <Nat>> (permutations <Nat> (fromto (0, 3))) (\ x =>
  System.IO.println (x); 
//  print System.IO.print x; write "\n";
  n := !n + 1
);
System.IO.println (!n)

end