let

open System.List

function foreach <a> (list : List <a>) : (a -> ()) -> () =
  fun (action : a -> ()) =>
    case list  of  Nil        =>  ()
               |   Cons cons  =>  action (cons.head) ; 
                                  foreach <a> (cons.tail) action
    end

function sum (list : List <Nat>) : Nat =
  let  val s = ref zero
  in   foreach <Nat> (list) 
         (fun (n : Nat) => s := add (!s, n)) ; 
       !s
  end

function map <a, b> (f : a -> b) : (List <a> -> List <b>) =
  fun (list : List <a>) =>
    case list  of  Nil        =>  Nil
               |   Cons cons  =>  Cons (  head  =  f (cons.head),
                                          tail  =  map <a, b> f (cons.tail) )
    end

function reduce <a, b> (a : b, op : (a, b) -> b) : List <a> -> b =
  fun (list : List <a>) =>
    case list  of  Nil        =>  a
               |   Cons cons  =>  
                     op (cons.head, reduce <a, b> (a, op) (cons.tail))
    end

val sum = reduce <Nat, Nat> (zero, add)

in
reduce <Nat, Nat> (succ zero, mul);
reduce <Bool, Bool> (false, \ (a, b) => a || b);
reduce <Bool, Bool> (true, \ (a, b) => a && b);
reduce <a, List <a>> (Nil, fun (x : a, xs : List <a>) => Cons (head = x, tail = xs));
// reduce <a, List <a>> (Nil, fun (x : a, xs : List <a>) => insert <a> (ord, x, xs));
0
end
